### EXAMPLE 1: demonstrates how measurement error can lead to a type sign error in a covariate
### What kind of data invalidates fong + tyler?
### Even when you have a good predictor, if it's biased against a covariate you can get the wrong sign.
### Even when you include the proxy variable in the regression.
### But with some ground truth and multiple imputation, you can fix it.

library(argparser)
library(mecor)
library(ggplot2)
library(data.table)
library(filelock)
library(arrow)
library(Amelia)
library(Zelig)
library(predictionError)
options(amelia.parallel="no",
        amelia.ncpus=1)
setDTthreads(40)

source("simulation_base.R")

## SETUP:
### we want to estimate x -> y; x is MAR
### we have x -> k; k -> w; x -> w is used to predict x via the model w.
### A realistic scenario is that we have an NLP model predicting something like "racial harassment" in social media comments
### The labels x are binary, but the model provides a continuous predictor

### simulation:
#### how much power do we get from the model in the first place? (sweeping N and m)
#### 

## one way to do it is by adding correlation to x.obs and y that isn't in w.
## in other words, the model is missing an important feature of x.obs that's related to y.
simulate_data <- function(N, m, B0, Bxy, Bgy, seed, y_explained_variance=0.025, prediction_accuracy=0.73, accuracy_imbalance_difference=0.3){
    set.seed(seed)
    # make w and y dependent
    g <- rbinom(N, 1, 0.5)
    x <- rbinom(N, 1, 0.5)

    y.var.epsilon <- (var(Bgy * g) + var(Bxy *x) + 2*cov(Bgy*g,Bxy*x)) * ((1-y_explained_variance)/y_explained_variance)
    y.epsilon <- rnorm(N, sd = sqrt(y.var.epsilon))
    y <- Bgy * g + Bxy * x + y.epsilon

    df <- data.table(x=x,y=y,g=g)

    if(m < N){
        df <- df[sample(nrow(df), m), x.obs := x]
    } else {
        df <- df[, x.obs := x]
    }

    df <- df[,w_pred:=x]

    pg <- mean(g)
    px <- mean(x)
    accuracy_imbalance_ratio <- (prediction_accuracy + accuracy_imbalance_difference/2) / (prediction_accuracy - accuracy_imbalance_difference/2)

    # this works because of conditional probability
    accuracy_g0 <- prediction_accuracy / (pg*(accuracy_imbalance_ratio) + (1-pg))
    accuracy_g1 <- accuracy_imbalance_ratio * accuracy_g0

    dfg0 <- df[g==0]
    ng0 <- nrow(dfg0)
    dfg1 <- df[g==1]
    ng1 <- nrow(dfg1)

    dfg0 <- dfg0[sample(ng0, (1-accuracy_g0)*ng0), w_pred := (w_pred-1)**2]
    dfg1 <- dfg1[sample(ng1, (1-accuracy_g1)*ng1), w_pred := (w_pred-1)**2]

    df <- rbind(dfg0,dfg1)

    w <- predict(glm(x ~ w_pred,data=df,family=binomial(link='logit')),type='response')
    df <- df[,':='(w=w, w_pred = w_pred)]
    return(df)
}

parser <- arg_parser("Simulate data and fit corrected models")
parser <- add_argument(parser, "--N", default=5000, help="number of observations of w")
parser <- add_argument(parser, "--m", default=200, help="m the number of ground truth observations")
parser <- add_argument(parser, "--seed", default=432, help='seed for the rng')
parser <- add_argument(parser, "--outfile", help='output file', default='example_2.feather')
parser <- add_argument(parser, "--y_explained_variance", help='what proportion of the variance of y can be explained?', default=0.01)
parser <- add_argument(parser, "--prediction_accuracy", help='how accurate is the predictive model?', default=0.73)
parser <- add_argument(parser, "--accuracy_imbalance_difference", help='how much more accurate is the predictive model for one class than the other?', default=0.3)

args <- parse_args(parser)

B0 <- 0
Bxy <- 0.2
Bgy <- -0.2

df <- simulate_data(args$N, args$m, B0, Bxy, Bgy, args$seed, args$y_explained_variance, args$prediction_accuracy, args$accuracy_imbalance_difference)

result <- list('N'=args$N,'m'=args$m,'B0'=B0,'Bxy'=Bxy,'Bgy'=Bgy, 'seed'=args$seed, 'y_explained_variance'=args$y_explained_variance, 'prediction_accuracy'=args$prediction_accuracy, 'accuracy_imbalance_difference'=args$accuracy_imbalance_difference)

outline <- run_simulation_depvar(df=df, result)


outfile_lock <- lock(paste0(args$outfile, '_lock'),exclusive=TRUE)
if(file.exists(args$outfile)){
    logdata <- read_feather(args$outfile)
    logdata <- rbind(logdata,as.data.table(outline))
} else {
    logdata <- as.data.table(outline)
}

print(outline)
write_feather(logdata, args$outfile)
unlock(outfile_lock)
